import pytest

MB = 2**20


def test_multipart_flow(base_app, s3fs):
    part_size = 7 * MB
    last_part_size = 5 * MB

    # initialize the upload
    upload_metadata = dict(
        parts=2, part_size=part_size, size=part_size + last_part_size
    )
    upload_metadata |= s3fs.multipart_initialize_upload(**upload_metadata) or {}

    # can not commit just now because no parts were uploaded
    with pytest.raises(ValueError):
        s3fs.multipart_commit_upload(**upload_metadata)

    # check that links are generated

    links = s3fs.multipart_links(**upload_metadata)["parts"]
    assert len(links) == 2
    assert links[0]["part"] == 1
    assert "url" in links[0]
    assert links[1]["part"] == 2
    assert "url" in links[1]

    # upload the first part manually
    multipart_file = s3fs.multipart_file(upload_metadata["uploadId"])
    multipart_file.upload_part(1, b"0" * part_size)
    assert len(multipart_file.get_parts(2)) == 1

    # still can not commit because not all parts were uploaded
    with pytest.raises(ValueError):
        s3fs.multipart_commit_upload(**upload_metadata)

    # upload the second part
    multipart_file.upload_part(2, b"1" * last_part_size)
    assert len(multipart_file.get_parts(2)) == 2

    s3fs.multipart_commit_upload(**upload_metadata)

    assert s3fs.open("rb").read() == b"0" * part_size + b"1" * last_part_size


def test_multipart_abort(base_app, s3fs):
    part_size = 7 * MB
    last_part_size = 5 * MB

    # initialize the upload
    upload_metadata = dict(
        parts=2, part_size=part_size, size=part_size + last_part_size
    )
    upload_metadata |= s3fs.multipart_initialize_upload(**upload_metadata) or {}

    s3fs.multipart_abort_upload(**upload_metadata)


def test_set_content_not_supported(base_app, s3fs):
    part_size = 7 * MB
    last_part_size = 5 * MB

    # initialize the upload
    upload_metadata = dict(
        parts=2, part_size=part_size, size=part_size + last_part_size
    )
    upload_metadata |= s3fs.multipart_initialize_upload(**upload_metadata) or {}

    with pytest.raises(NotImplementedError):
        s3fs.multipart_set_content(1, b"0" * part_size, part_size, **upload_metadata)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019 Esteban J. G. Gabancho.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Pytest configuration."""
import hashlib

import boto3
import pytest
from flask import current_app
from invenio_app.factory import create_api
from moto import mock_aws

from invenio_s3 import S3FSFileStorage


@pytest.fixture(scope="module")
def app_config(app_config):
    """Customize application configuration."""
    app_config["FILES_REST_STORAGE_FACTORY"] = "invenio_s3.s3fs_storage_factory"
    app_config["S3_ENDPOINT_URL"] = None
    app_config["S3_ACCESS_KEY_ID"] = "test"
    app_config["S3_SECRECT_ACCESS_KEY"] = "test"
    app_config["THEME_FRONTPAGE"] = False
    return app_config


@pytest.fixture(scope="module")
def create_app():
    """Application factory fixture."""
    return create_api


@pytest.fixture(scope="function")
def s3_bucket(appctx):
    """S3 bucket fixture."""
    with mock_aws():
        session = boto3.Session(
            aws_access_key_id=current_app.config.get("S3_ACCESS_KEY_ID"),
            aws_secret_access_key=current_app.config.get("S3_SECRECT_ACCESS_KEY"),
        )
        s3 = session.resource("s3")
        bucket = s3.create_bucket(Bucket="test_invenio_s3")

        yield bucket

        for obj in bucket.objects.all():
            obj.delete()
        bucket.delete()


@pytest.fixture(scope="function")
def s3fs_testpath(s3_bucket):
    """S3 test path."""
    return "s3://{}/path/to/data".format(s3_bucket.name)


@pytest.fixture(scope="function")
def s3fs(s3_bucket, s3fs_testpath):
    """Instance of S3FSFileStorage."""
    s3_storage = S3FSFileStorage(s3fs_testpath)
    return s3_storage


@pytest.fixture
def file_instance_mock(s3fs_testpath):
    """Mock of a file instance."""

    class FileInstance(object):
        def __init__(self, **kwargs):
            for k, v in kwargs.items():
                setattr(self, k, v)

    return FileInstance(
        id="deadbeef-65bd-4d9b-93e2-ec88cc59aec5",
        uri=s3fs_testpath,
        size=4,
        updated=None,
    )


@pytest.fixture()
def get_md5():
    """Get MD5 of data."""

    def inner(data, prefix=True):
        m = hashlib.md5()
        m.update(data)
        return "md5:{0}".format(m.hexdigest()) if prefix else m.hexdigest()

    return inner



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019 Esteban J. G. Gabancho.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""File serving helpers for S3 files."""

import mimetypes
import unicodedata
from urllib.parse import quote

from flask import current_app
from invenio_files_rest.helpers import sanitize_mimetype
from werkzeug.datastructures import Headers


def redirect_stream(
    s3_url_builder,
    filename,
    mimetype=None,
    restricted=True,
    as_attachment=False,
    trusted=False,
):
    """Redirect to URL to serve the file directly from there.

    :param url: redirection URL

    :return: Flask response.
    """
    # Guess mimetype from filename if not provided.
    if mimetype is None and filename:
        mimetype = mimetypes.guess_type(filename)[0]
    if mimetype is None:
        mimetype = "application/octet-stream"

    # Construct headers
    headers = Headers()

    if not trusted:
        # Sanitize MIME type
        mimetype = sanitize_mimetype(mimetype, filename=filename)
        # See https://www.owasp.org/index.php/OWASP_Secure_Headers_Project
        # Prevent JavaScript execution
        headers["Content-Security-Policy"] = "default-src 'none';"
        # Prevent MIME type sniffing for browser.
        headers["X-Content-Type-Options"] = "nosniff"
        # Prevent opening of downloaded file by IE
        headers["X-Download-Options"] = "noopen"
        # Prevent cross domain requests from Flash/Acrobat.
        headers["X-Permitted-Cross-Domain-Policies"] = "none"
        # Prevent files from being embedded in frame, iframe and object tags.
        headers["X-Frame-Options"] = "deny"
        # Enable XSS protection (IE, Chrome, Safari)
        headers["X-XSS-Protection"] = "1; mode=block"

    # Force Content-Disposition for application/octet-stream to prevent
    # Content-Type sniffing.
    if as_attachment or mimetype == "application/octet-stream":
        # see https://github.com/pallets/werkzeug/blob/main/src/werkzeug/utils.py#L456-L465
        try:
            filename.encode("ascii")
        except UnicodeEncodeError:
            simple = unicodedata.normalize("NFKD", filename)
            simple = simple.encode("ascii", "ignore").decode("ascii")
            # safe = RFC 5987 attr-char
            quoted = quote(filename, safe="!#$&+-.^_`|~")
            filenames = {"filename": simple, "filename*": f"UTF-8''{quoted}"}
        else:
            filenames = {"filename": filename}
        headers.add("Content-Disposition", "attachment", **filenames)
    else:
        headers.add("Content-Disposition", "inline")

    url = s3_url_builder(
        ResponseContentType=mimetype,
        ResponseContentDisposition=headers.get("Content-Disposition"),
    )
    headers["Location"] = url

    # TODO: Set cache-control
    # if not restricted:
    #     rv.cache_control.public = True
    #     cache_timeout = current_app.get_send_file_max_age(filename)
    #     if cache_timeout is not None:
    #         rv.cache_control.max_age = cache_timeout
    #         rv.expires = int(time() + cache_timeout)
    # Construct response object.

    rv = current_app.response_class(
        url,
        status=302,
        headers=headers,
        mimetype=mimetype,
        direct_passthrough=True,
    )

    return rv



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018, 2019, 2020, 2021 Esteban J. G. Gabancho.
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""S3 file storage interface."""

from functools import partial, wraps
from math import ceil

import s3fs
from flask import current_app
from invenio_files_rest.errors import StorageError
from invenio_files_rest.storage import PyFSFileStorage, pyfs_storage_factory

from .helpers import redirect_stream
from .multipart_client import MultipartS3File


def set_blocksize(f):
    """Decorator to set the correct block size according to file size."""

    @wraps(f)
    def inner(self, *args, **kwargs):
        size = kwargs.get("size", None)
        block_size = (
            ceil(size / current_app.config["S3_MAXIMUM_NUMBER_OF_PARTS"])
            if size
            else current_app.config["S3_DEFAULT_BLOCK_SIZE"]
        )

        if block_size > self.block_size:
            self.block_size = block_size
        return f(self, *args, **kwargs)

    return inner


class S3FSFileStorage(PyFSFileStorage):
    """File system storage using Amazon S3 API for accessing files."""

    def __init__(self, fileurl, **kwargs):
        """Storage initialization."""
        self.block_size = current_app.config["S3_DEFAULT_BLOCK_SIZE"]
        super(S3FSFileStorage, self).__init__(fileurl, **kwargs)

    def _get_fs(self, *args, **kwargs):
        """Get PyFilesystem instance and S3 real path."""
        if not self.fileurl.startswith("s3://"):
            return super(S3FSFileStorage, self)._get_fs(*args, **kwargs)

        info = current_app.extensions["invenio-s3"].init_s3fs_info
        fs = s3fs.S3FileSystem(default_block_size=self.block_size, **info)

        return (fs, self.fileurl)

    @set_blocksize
    def initialize(self, size=0):
        """Initialize file on storage and truncate to given size."""
        fs, path = self._get_fs()

        if fs.exists(path):
            fp = fs.rm(path)
        fp = fs.open(path, mode="wb")

        try:
            to_write = size
            fs_chunk_size = fp.blocksize  # Force write every time
            while to_write > 0:
                current_chunk_size = (
                    to_write if to_write <= fs_chunk_size else fs_chunk_size
                )
                fp.write(b"\0" * current_chunk_size)
                to_write -= current_chunk_size
        except Exception:
            fp.close()
            self.delete()
            raise
        finally:
            fp.close()

        self._size = size

        return self.fileurl, size, None

    def delete(self):
        """Delete a file."""
        fs, path = self._get_fs()
        if fs.exists(path):
            fs.rm(path)
        return True

    @set_blocksize
    def update(
        self,
        incoming_stream,
        seek=0,
        size=None,
        chunk_size=None,
        progress_callback=None,
    ):
        """Update a file in the file system."""
        old_fp = self.open(mode="rb")
        updated_fp = S3FSFileStorage(self.fileurl, size=self._size).open(mode="wb")
        try:
            if seek >= 0:
                to_write = seek
                fs_chunk_size = updated_fp.blocksize
                while to_write > 0:
                    current_chunk_size = (
                        to_write if to_write <= fs_chunk_size else fs_chunk_size
                    )
                    updated_fp.write(old_fp.read(current_chunk_size))
                    to_write -= current_chunk_size

            bytes_written, checksum = self._write_stream(
                incoming_stream,
                updated_fp,
                chunk_size=chunk_size,
                size=size,
                progress_callback=progress_callback,
            )

            if (bytes_written + seek) < self._size:
                old_fp.seek((bytes_written + seek))
                to_write = self._size - (bytes_written + seek)
                fs_chunk_size = updated_fp.blocksize
                while to_write > 0:
                    current_chunk_size = (
                        to_write if to_write <= fs_chunk_size else fs_chunk_size
                    )
                    updated_fp.write(old_fp.read(current_chunk_size))
                    to_write -= current_chunk_size
        finally:
            old_fp.close()
            updated_fp.close()

        return bytes_written, checksum

    def send_file(
        self,
        filename,
        mimetype=None,
        restricted=True,
        checksum=None,
        trusted=False,
        chunk_size=None,
        as_attachment=False,
    ):
        """Send the file to the client."""
        try:
            fs, path = self._get_fs()
            s3_url_builder = partial(
                fs.url, path, expires=current_app.config["S3_URL_EXPIRATION"]
            )

            return redirect_stream(
                s3_url_builder,
                filename,
                mimetype=mimetype,
                restricted=restricted,
                trusted=trusted,
                as_attachment=as_attachment,
            )
        except Exception as e:
            raise StorageError("Could not send file: {}".format(e))

    @set_blocksize
    def copy(self, src, *args, **kwargs):
        """Copy data from another file instance.

        If the source is an S3 stored object the copy process happens on the S3
        server side, otherwise we use the normal ``FileStorage`` copy method.
        """
        if src.fileurl.startswith("s3://"):
            fs, path = self._get_fs()
            fs.copy(src.fileurl, path)
        else:
            super(S3FSFileStorage, self).copy(src, *args, **kwargs)

    @set_blocksize
    def save(self, *args, **kwargs):
        """Save incoming stream to storage.

        Just overwrite parent method to allow set the correct block size.
        """
        return super(S3FSFileStorage, self).save(*args, **kwargs)

    def multipart_initialize_upload(self, parts, size, part_size):
        """
        Initialize a multipart upload.

        :param parts: The number of parts that will be uploaded.
        :param size: The total size of the file.
        :param part_size: The size of each part except the last one.
        :returns: a dictionary of additional metadata that should be stored between
            the initialization and the commit of the upload.
        """
        return {"uploadId": self.multipart_file().create_multipart_upload()}

    def multipart_file(self, upload_id=None):
        """Get a low-level file object.

        :param upload_id: The upload ID of the multipart upload, can be none to get a new upload.
        :returns: an instance of MultipartS3File.
        """
        # WARNING: low-level code. The underlying s3fs currently does not have support
        # for multipart uploads without keeping the S3File instance in memory between requests.
        return MultipartS3File(*self._get_fs(), upload_id=upload_id)

    def multipart_set_content(self, part, stream, content_length, **multipart_metadata):
        """Set the content of a part of the multipart upload.

        This method will never be called
        by invenio as user will use direct pre-signed requests to S3 and will never
        upload the files through Invenio.
        """
        raise NotImplementedError(
            "The multipart_set_content method is not implemented as it will never be called directly."
        )

    def multipart_commit_upload(self, **multipart_metadata):
        """Commit the multipart upload.

        :param multipart_metadata: The metadata returned by the multipart_initialize_upload
            and the metadata returned by the multipart_set_content for each part.
        :returns: The Etag of the completed upload.
        """
        f = self.multipart_file(multipart_metadata["uploadId"])
        expected_parts = int(multipart_metadata["parts"])
        parts = f.get_parts(max_parts=expected_parts)
        if len(parts) != expected_parts:
            raise ValueError(
                f"Not all parts were uploaded, got {len(parts)} out of {expected_parts} parts."
            )
        completed_response = f.complete_multipart_upload(parts)
        # in S3, the Etag is the md5 checksum of the md5checksums of the parts
        return completed_response.get("ETag", None).strip('"')

    def multipart_abort_upload(self, **multipart_metadata):
        """Abort the multipart upload.

        :param multipart_metadata: The metadata returned by the multipart_initialize_upload
            and the metadata returned by the multipart_set_content for each part.
        """
        f = self.multipart_file(multipart_metadata["uploadId"])
        f.abort_multipart_upload()

    def multipart_links(self, **multipart_metadata):
        """Generate links for the parts of the multipart upload.

        :param multipart_metadata: The metadata returned by the multipart_initialize_upload
            and the metadata returned by the multipart_set_content for each part.
        :returns: a dictionary of name of the link to invenio_records_resources.services.base.links.Link
        """
        return self.multipart_file(multipart_metadata["uploadId"]).get_part_links(
            int(multipart_metadata["parts"]),
            current_app.config["S3_UPLOAD_URL_EXPIRATION"],
        )


def s3fs_storage_factory(**kwargs):
    """File storage factory for S3."""
    return pyfs_storage_factory(filestorage_class=S3FSFileStorage, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Miroslav Simek
#
# Invenio-S3 is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Low level client for S3 multipart uploads."""

from datetime import datetime, timedelta, timezone

# WARNING: low-level code. The underlying s3fs currently does not have support
# for multipart uploads without keeping the S3File instance in memory between requests.
# To overcome this limitation, we have to use the low-level API directly separated in the
# MultipartS3File class.


class MultipartS3File:
    """Low level client for S3 multipart uploads."""

    def __init__(self, fs, path, upload_id=None):
        """Initialize the low level client.

        :param fs: S3FS instance
        :param path: The path of the file (with bucket and version)
        :param upload_id: The upload ID of the multipart upload, can be none to get a new upload.
        """
        self.fs = fs
        self.path = path
        self.bucket, self.key, self.version_id = fs.split_path(path)
        self.s3_client = fs.s3
        self.acl = fs.s3_additional_kwargs.get("ACL", "")
        self.upload_id = upload_id

    def create_multipart_upload(self):
        """Create a new multipart upload.

        :returns: The upload ID of the multipart upload.
        """
        mpu = self.s3_client.create_multipart_upload(
            Bucket=self.bucket, Key=self.key, ACL=self.acl
        )
        # TODO: error handling here
        self.upload_id = mpu["UploadId"]
        return self.upload_id

    def get_parts(self, max_parts):
        """List the parts of the multipart upload.

        :param max_parts: The maximum number of parts to list.
        :returns: The list of parts, including checksums and etags.
        """
        ret = self.s3_client.list_parts(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
            MaxParts=max_parts,
            PartNumberMarker=0,
        )
        return ret.get("Parts", [])

    def upload_part(self, part_number, data):
        """Upload a part of the multipart upload. Will be used only in tests.

        :param part_number: The part number.
        :param data: The data to upload.
        """
        part = self.s3_client.upload_part(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
            PartNumber=part_number,
            Body=data,
        )
        return part

    def _complete_operation_part_parameters(self, part):
        """Filter parameters for the complete operation."""
        ret = {}
        for k in [
            "PartNumber",
            "ETag",
            "ChecksumCRC32",
            "ChecksumCRC32C",
            "ChecksumSHA1",
            "ChecksumSHA256",
        ]:
            if k in part:
                ret[k] = part[k]
        return ret

    def get_part_links(self, max_parts, url_expiration):
        """Generate pre-signed URLs for the parts of the multipart upload.

        :param max_parts: The maximum number of parts to list.
        :param url_expiration: The expiration time of the URLs in seconds

        :returns: The list of parts with pre-signed URLs and expiration times.
        """
        # AWS S3 requires the expiration timestamp to be in UTC.
        expiration = datetime.now(timezone.utc) + timedelta(seconds=url_expiration)

        # AWS expects the timestamp in ISO 8601 format ending with 'Z' (indicating UTC),
        # e.g., '2025-05-20T16:03:11Z'.
        # Since `isoformat()` on a timezone-aware datetime includes '+00:00' instead of 'Z',
        # we first remove the timezone info and manually append 'Z'.
        expiration = expiration.replace(tzinfo=None).isoformat(timespec="seconds") + "Z"

        return {
            "parts": [
                {
                    "part": part + 1,
                    "url": self.s3_client.generate_presigned_url(
                        "upload_part",
                        Params={
                            "Bucket": self.bucket,
                            "Key": self.key,
                            "UploadId": self.upload_id,
                            "PartNumber": part + 1,
                        },
                        ExpiresIn=url_expiration,
                    ),
                    "expiration": expiration,
                }
                for part in range(max_parts)
            ]
        }

    def complete_multipart_upload(self, parts):
        """Complete the multipart upload.

        :param parts: The list of parts (as from self.get_parts), including checksums and etags.
        """
        return self.s3_client.complete_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
            MultipartUpload={
                "Parts": [
                    self._complete_operation_part_parameters(part) for part in parts
                ]
            },
        )

    def abort_multipart_upload(self):
        """Abort the multipart upload."""
        return self.s3_client.abort_multipart_upload(
            Bucket=self.bucket, Key=self.key, UploadId=self.upload_id
        )
